from pathlib import Path
from copy import copy

from Bio.PDB.Polypeptide import index_to_one, three_to_index

from .database_parser import DatabaseParser
from .file_handler import FileHandler
from .isoform import Isoform


class Gene:
    def __init__(self, gene_mutation_block: list[list], out_path: str):
        self.name = gene_mutation_block[0].upper()
        self.mutations = self._standardize_mutations(gene_mutation_block[1:])
        self.mappable_mutations = []
        self.out_path = Path(out_path, self.name)
        self.create_directory()
        self.sequences = []
        self.isoforms = []

        self.isoforms_to_model = []

    def create_directory(self) -> None:
        """Create an empty directory with the gene name
        at the output path specified by the user."""

        with FileHandler() as fh:
            fh.create_directory(self.out_path)

    def load_sequences(self, db_parser: DatabaseParser) -> None:
        """Take the sequences from the Uniprot databases and
        save them as attributes

        :param db_parser: instance of DatabaseParser
        """
        canonical_sequences = db_parser.get_canonical_isoforms(self.name)
        self.sequences = canonical_sequences
        noncanonical_sequences = db_parser.get_noncanonical_isoforms(self.name)
        self.sequences += noncanonical_sequences

    def load_isoforms(self, db_parser: DatabaseParser) -> None:
        """Create an Isoform instance, for each sequence
        found in the databases, if at least a mutation can
         be mapped onto it.

        :param db_parser: instance of DatabaseParser.
        """
        self.load_sequences(db_parser)
        for isoform_index, sequence in enumerate(self.sequences):

            # Make a list of all the mutations that can be directly mapped
            # onto this isoform
            modellable_mutations = []
            # This makes the multiline fasta into one-line:
            sequence_string = "".join("".join(sequence[1:]).splitlines())
            for mutation in self.mutations:
                if self._check_presence_mutated_residue(
                    sequence_string, mutation
                ):
                    # The mutation can be mapped, append it to the list.
                    # The list will be passed to the Isoform object.
                    modellable_mutations.append(mutation)

                    # Take note of the mutations that can be mapped on
                    # at least on isoform. This will be necessary to
                    # calculate the score of the mutation function.
                    # See doi: 10.3389/fchem.2022.1059593

                    if mutation not in self.mappable_mutations:
                        self.mappable_mutations.append(mutation)

            # Check if no mutations can be mapped. Skip isoform.
            if len(modellable_mutations) == 0:
                print(
                    "None of the mutations can be mapped on"
                    f" {self.name} isoform_{isoform_index}"
                )

            else:
                self.isoforms.append(
                    Isoform(
                        self.name,
                        sequence,
                        isoform_index,
                        self.out_path,
                        modellable_mutations,
                    )
                )

    def _check_presence_mutated_residue(
        self, sequence: str, mutation: list
    ) -> bool:
        """Given a mutation, check if that specific residue is present at
        that position.

        Args:
            sequence (str): The sequence of the isoform expressed as
            a one-line string.
            mutation (list): The mutation expressed as a list, with
            one-letter residue names (e.g., ['R','899','A'])

        Returns:
            can_be_mutated (bool): True if mutation can be applied
            without changing residue number.
        """
        residue_ID=int(mutation[1]) - 1

        if len(sequence)<residue_ID+1: #if sequence too short
            return False

        can_be_mutated = sequence[residue_ID] == mutation[0]
        return can_be_mutated

    def _standardize_mutations(self, mutation_list) -> list:
        """Remove white spaces from mutations. Change residue names
        to one-letter format.
        Args:
            mutation_list (list): list of mutations as obtained
            from the input file.

        Returns:
            standard_mutation_list (list): standardized list of mutations.
        """
        standard_mutation_list = []
        standard_residues = [
            "ALA",
            "ARG",
            "ASN",
            "ASP",
            "CYS",
            "GLU",
            "GLN",
            "GLY",
            "HIS",
            "ILE",
            "LEU",
            "LYS",
            "MET",
            "PHE",
            "PRO",
            "SER",
            "THR",
            "TRP",
            "TYR",
            "VAL",
        ]

        for i, mutation in enumerate(mutation_list):
            mutation = "".join(mutation.split())  # remove whitespaces

            first_resname = []
            number = []

            letter_index = 0
            while mutation[letter_index].isalpha():
                first_resname.append(mutation[letter_index])
                letter_index += 1

            while mutation[letter_index].isnumeric():
                number.append(mutation[letter_index])
                letter_index += 1

            second_resname = mutation[letter_index:]

            # Apply checks on resiude names
            first_resname = "".join(first_resname).upper()
            second_resname = "".join(second_resname).upper()

            if len(first_resname) == len(second_resname):
                if len(first_resname) == 3:
                    if (first_resname in standard_residues) and (
                        second_resname in standard_residues
                    ):
                        first_resname = index_to_one(
                            three_to_index(first_resname)
                        )
                        second_resname = index_to_one(
                            three_to_index(second_resname)
                        )
                    else:
                        raise (
                            ValueError(
                                f"Residue names in {mutation} do not"
                                " appear to be standard."
                            )
                        )
                else:
                    if len(first_resname) != 1:
                        raise (
                            ValueError(
                                f"Residue names in {mutation} must be in"
                                "one-letter or three-letter format."
                            )
                        )

            else:
                raise (
                    ValueError(
                        "Different residue name format"
                        f" in mutation {mutation}."
                    )
                )

            standard_mutation_list.append(
                [
                    "".join(first_resname),
                    "".join(number),
                    "".join(second_resname),
                ]
            )

        return standard_mutation_list

    def report_on_selected_isoforms(self):
        """Print on screen a report on the selected isoforms."""

        print(
            f"For the gene {self.name} the following models will be created,\n"
            "if the mutation is covered by the strctural templates:"
        )
        for modellable_isoform in self.isoforms_to_model:
            print(
                f"{modellable_isoform[0].isoform_name} "
                + "".join(modellable_isoform[1])
            )

    def select_isoforms(
        self, w1: float, w2: float, sequence_identity_cutoff: float,
        model_cutoff: int,cobalt_home
    ) -> None:
        """Start the calculation of the scores for all isoforms. The
        templates of each isoform are filtered, based on the sequence
        identity. The isoforms that still have templates
        are ordered by their selection score.

        Args:
            w1 (float): weight of the structural function
            w2 (float): weight of the mutation function
            sequence_identity_cutoff (float): templates with a seq. identity
                                              lower than this, are excluded.
        """
        for isoform in self.isoforms:
            isoform.calculate_score(self.mappable_mutations,model_cutoff,
                                    sequence_identity_cutoff,w1,w2,
                                    cobalt_home)

        self.isoforms = [
            isoform for isoform in self.isoforms if isoform.modellable
        ]
        if len(self.isoforms) == 0:
            print(f"No modellable isoform for gene {self.name}")
        else:
            self.isoforms.sort(key=lambda x: -x.selection_score)

            # Add wild type to list of isoforms to model
            self.isoforms_to_model.append([copy(self.isoforms[0]), "WT"])

            # Take note of mutations to model
            mutations_to_model = self.mutations[:]

            # Add mutations of best isoform
            for mutation in self.isoforms[0].mutations:
                self.isoforms_to_model.append([copy(self.isoforms[0]), mutation])
                mutations_to_model.remove(mutation)

            # Try other isoforms to model all mutations
            # if there are other isoforms to select from
            if len(self.isoforms) > 0:
                for mutation in mutations_to_model:
                    check_across_isoforms = [
                        mutation in isoform.mutations
                        for isoform in self.isoforms
                    ]
                    if sum(check_across_isoforms) > 0:
                        isoform_for_mutation = self.isoforms[
                            check_across_isoforms.index(True)
                        ]
                        self.isoforms_to_model.append(
                            [isoform_for_mutation, mutation]
                        )
                        mutations_to_model.remove(mutation)

            # Print the mutations with that cannot
            # be associated to any isoform
            if len(mutations_to_model) > 0:
                print(
                    "The following mutations will not be"
                    " modelled for gene " + self.name
                )
                print(",".join(["".join(mut) for mut in mutations_to_model]))

    def write_report(self):
        """ For every isoform and modeller run,
        append their information to a report.md file
        contained in the gene directory.
        """
        
        content="# LOG FILE\n"
        content+="- GENE NAME: "+self.name+"\n"
        content+="- PATH: "+str(self.out_path.absolute())+"\n"
        content+="- REQUESTED MUTATIONS:\n"
        
        for mutation in self.mutations:
            content+="-- "+"".join(mutation)+"\n"

        content+="- MAPPABLE MUTATIONS:\n"
        
        for mutation in self.mappable_mutations:
            content+="-- "+"".join(mutation)+"\n"
            
        content+="- ISOFORM SCORES:\n"
        
        for isoform in self.isoforms:
            content+="-- "+isoform.isoform_name+"\n"
            content+="--- Mutation:"
            content+=str(round(isoform.mutation_score,2))+" Structural:"
            content+=str(round(isoform.structural_score,2))+" Selection:"
            content+=str(round(isoform.selection_score,2))+"\n"
            content+="--- Modellable: "+str(isoform.modellable)+"\n"
            
        #Add a list of mutations excluded due missing structural converage
        non_covered_isoforms=[]
        for isoform in self.isoforms_to_model:
            if isoform[0].modeller_run.mutation_is_modellable==False:
                non_covered_isoforms.append(isoform)
        if len(non_covered_isoforms)>0:
            content+="\nThe following mutations are in a region not covered by templates:\n"
            for isoform in non_covered_isoforms:
                content+="".join(isoform[1])+"\n"
        
        content+="\n## MODELS\n"
        
        for isoform in self.isoforms_to_model:
            #isoform[0] is the object, isoform[1] is the mutation
            content+="- "+isoform[0].isoform_name+" "+"".join(isoform[1])+"\n"
            content+="- Templates: "
            
            for template in isoform[0].templates:
                content+=template.pdb_name+"_"+template.pdb_chain+" "
            
            content+="\n"
            
            content+="\n".join(isoform[0].modeller_run.logged_scores)+"\n\n"
            
    
        with FileHandler() as fh:
            fh.write_file(Path(self.out_path,"report.md"),content)


import os
import subprocess
from pathlib import Path

from .file_handler import FileHandler


class Modeller_manager:
    def __init__(self, isoform, mutation: list, modeller_exec: str, model_cutoff: int,
                 number_of_wt:int,number_of_mut:int):
        self.isoform = isoform
        self.mutation = mutation
        self.sequence_to_model = self.isoform.aligned_sequence[:]
        self.modeller_exec = modeller_exec
        self.model_cutoff = model_cutoff
        self.logged_scores=[]
        self.num_chains=1
        self.chain_starts=[]
        self.mutation_is_modellable=True
        
        if "".join(mutation)=="WT":
            self.num_models=number_of_wt
        else:
            self.num_models=number_of_mut
    def write(self):

        print(
            f"Modelling {self.isoform.gene_name}"
            f" {self.isoform.isoform_name} "+"".join(self.mutation)
        )
        self.write_alignment()
        self.write_script()
    
    def _check_modellability(self,sequence) -> bool:
        """Sometimes the mutation can be mapped on the isoform, but it is not
        in the range of the protein that is covered by the templates. The Gene
        object that starts the modelling process, should be updated to stop the
        modelling, and write and informed report.
        
        This function checks the mutation residue number in the sequence and returns
        a boolean representing the presence of the mutated residue. 
        
        Args:
        sequences (list): A list of [residue number, residue], with one element
        for each residue of the target sequence, after adding chain breaks.
        
        Returns:
        bool: Presence of the mutated residue in the final alignment."""
        
        if self.mutation=='WT':
            return True
        
        resnum_to_mutate=int(self.mutation[1])
        resnumbers=[r[0] for r in sequence]
        self.mutation_is_modellable = resnum_to_mutate in resnumbers
        
        if not self.mutation_is_modellable:
            self.isoform.mutations_not_in_structure.append(self.mutation)
        
        return self.mutation_is_modellable

    def write_script(self):
        alignment_name = "../modeller_input_" + "".join(self.mutation) + ".dat"
        output_name = "../modeller_output_" + "".join(self.mutation) + ".dat"
        script_path = Path(
            self.isoform.out_path,
            "run_modeller_" + "".join(self.mutation) + ".py",
        )
        template_names = [
            t.pdb_name + "_" + t.pdb_chain + "_clean"
            for t in self.isoform.templates
        ]
        content = (
            """from modeller import *
from modeller.automodel import *
from modeller.scripts import complete_pdb

class MyModel(AutoModel):
    def special_patches(self, aln):
        # Rename both chains and renumber the residues in each
        self.rename_segments(segment_ids="""+str(["A" for  i in range(self.num_chains)])+""",
                             renumber_residues="""+str(self.chain_starts)+""")

log.verbose()
env = environ()
env.io.atom_files_directory = '../templates/'
env.io.hetatm = True
a = MyModel(env,
    alnfile =\""""
            + alignment_name
            + """\",
    knowns = ("""
            + str(template_names)
            + """),
    sequence = \""""
            + self.isoform.gene_name
            + """\",
    assess_methods=(assess.DOPE, assess.GA341))
a.starting_model= 1
a.ending_model  = """+str(self.num_models)+"""
a.make()
ok_models = filter(lambda x: x['failure'] is None, a.outputs)
toscore = 'DOPE score'

models= list(ok_models)
models.sort(lambda k: k[toscore])

myout = open(\""""
            + output_name
            + """\", "w")
for m in models:
        myout.write(str(m['name']) + " (DOPE SCORE: %.3f)" % (m[toscore]))
env.libs.topology.read(file='$(LIB)/top_heav.lib')
env.libs.parameters.read(file='$(LIB)/par.lib')
mdl = complete_pdb(env, m['name'])
s = selection(mdl)
s.assess_dope(output='ENERGY_PROFILE NO_REPORT', file=\""""
            + self.isoform.gene_name
            + """\", normalize_profile=True, smoothing_window=15)"""
        )

        with FileHandler() as fh:
            fh.write_file(script_path, content)

    def _mutate_residue(self, mutation, sequence:str) -> bool:
        """Take a mutation in the format
        [1 letter amino acid,residue number, 1 lett. amino acid]
        and apply it to the aligned sequence to model,
        taking into account all the "-"'s.

        Returns edited sequence
        """
 
        if mutation=="WT":
            return sequence
        

        i = int(mutation[1]) - 1
        while i < len(sequence):
            actual_residue_index = i - sequence[:i].count("-")
            if (
                actual_residue_index + 1 == int(mutation[1])
            ) and sequence[i] == mutation[0]:
                sequence = (
                    sequence[:i]
                    + mutation[2]
                    + sequence[i + 1 :]
                )
                return sequence
            i += 1

        print(f"Could not apply mutation {mutation}!")
        return sequence

    def _add_number_to_sequence(self,sequence:str)->list:
        """From a sequence, return a 2D list with [resnum,residue]"""
        numbered_seq=[]
        i=1
        for char in sequence:
            if (char=="-"):
                numbered_seq.append(["-",char])
            else:
                numbered_seq.append([i,char])
                i+=1
        return numbered_seq
    
    def _save_chain_starts(self,sequence_with_resnum:list)->None:
        """Given a sequence saved as a list [resnum, residue],
        save all the residue numbers that represent an interruption
        in the continuity of the protein chain."""

        #lists al resnums without gaps
        filtered_numbers=[r[0] for r in sequence_with_resnum if r[0]!="-"]

   
        self.chain_starts.append(filtered_numbers[0])

        for i in range(1,len(filtered_numbers)):
            if filtered_numbers[i]!=filtered_numbers[i-1]+1:
                self.chain_starts.append(filtered_numbers[i]+1)
    
    def _add_chain_breaks(self, sequences: list) -> list:
        """For alignments in which there is no coverage for self.model_cutoff+
        residues, the section without coverage is replaced by
        chain breaks.

        Args:
            sequences (list): A list of [name, sequence] lists,
            one for each sequence to be written in the modeller
            alignment input file.

        Returns:
            list: Returns the same list, with chain breaks.
        """

        # Separate protein names and aligned sequences in two objects
        names = [object[0] for object in sequences]
        aligned_seq = [object[1] for object in sequences]

        #Add object to keep trak of residue number in target seq
        num_target_seq=self._add_number_to_sequence(aligned_seq[0])
        # Calculate worst case number of residues without coverage
        max_non_covered = max([seq.count("-") for seq in aligned_seq])

        # Search for non-covered subsequences with lenght between
        # max_non_covered and model_cutoff
        while max_non_covered >= self.model_cutoff:

            # check if "-" repeated max_non_covered times
            # is present in all the aligned structures (except target seqence)
            check_presence = [
                "-" * max_non_covered in seq for seq in aligned_seq[1:]
            ]
            if sum(check_presence) == (len(aligned_seq) - 1):

                # Find position of the sequence of "-"'s in all
                # aligned sequences for each position it will
                # check if the coverage is missing in all the
                # sequences
                positions = [
                    seq.find("-" * max_non_covered) for seq in aligned_seq[1:]
                ]
                for pos in positions:
                    check_position = [
                        seq[pos : pos + max_non_covered]
                        == "-" * max_non_covered
                        for seq in aligned_seq[1:]
                    ]
                    if sum(check_position) == (len(aligned_seq) - 1):
                        # If that is the case, replace that section in
                        # all sequences (+ target sequences) with a chain
                        # break ("/").
                        num_target_seq=num_target_seq[:pos]+num_target_seq[pos + max_non_covered -1:]
                        for i, seq in enumerate(aligned_seq):
                            aligned_seq[i] = (
                                seq[:pos] + "/" + seq[pos + max_non_covered :])
                            self.num_chains+=1

                        max_non_covered += (
                            1  # This repeats the check for the current value
                        )
            max_non_covered -= 1

        sequences = [[names[i], aligned_seq[i]] for i in range(len(names))]
        
        #Remove "/" at the beginning or end of the sequences:
        if sequences[0][1][0]=="/":
            sequences=[[s[0],s[1][1:]] for s in sequences]
        if sequences[0][1][-1]=="/":
            sequences=[[s[0],s[1][:-1]] for s in sequences]
        #Next line is to save resnum at chain breaks as local attributes
        self._save_chain_starts(num_target_seq)
        
        #Next line is to check if mutation can still be mapped
        _ =self._check_modellability(num_target_seq)
        
        return sequences

    def write_alignment(self):

        alignment_name = "modeller_input_" + "".join(self.mutation) + ".dat"
        output_path = Path(self.isoform.out_path, alignment_name)
        # Create an object storing all names and aligned sequences
        sequences = []
        with FileHandler() as fh:
            file_path=Path(self.isoform.out_path,"templates_aligned.fasta")
            aligned_sequence_file=fh.read_file(file_path)
            aligned_sequences=aligned_sequence_file.split(">")[1:]
            sequences=[[x.split("\n")[0].split()[-1],"".join(x.split('\n')[1:])] for x in aligned_sequences]

        #Add mutation if needed
        sequences[0][1]=self._mutate_residue(self.mutation,sequences[0][1])
        
        # Add chain breaks in place of long seqs with no coverage
        sequences = self._add_chain_breaks(sequences)


        # Start writing the content string to be printed in the file
        content = ""
        content += ">P1;" + self.isoform.gene_name + "\n"
        content += "sequence:" + self.isoform.gene_name + ":.:.:.:.::::\n"
        content += sequences[0][1] + "*\n"

        # Add templates
        for template_sequence in sequences[1:]:
            content += ">P1;" + template_sequence[0] + "_clean" + "\n"
            content += (
                "structureX:"
                + template_sequence[0]
                + "_clean"
                + ":.:.:.:.::::\n"
            )
            content += template_sequence[1] + "*\n"

        #After the last alignment it is know if the mutation
        #Can be certainly applied. Apply last check before
        #Writing
        if self.mutation_is_modellable:
            with FileHandler() as fh:
                fh.write_file(output_path, content)

    def run(self) -> None:
        
        if self.mutation_is_modellable==False:
            print("Modelling of mutation "+"".join(self.mutation)+
                  " was excluded on "+self.isoform.gene_name+" "+
                  self.isoform.isoform_name+" due to missing coverage.")
            self.load_log_file()
            return None
        
        model_path = Path(
            self.isoform.out_path,
            self.isoform.gene_name + "_" + "".join(self.mutation) + "_model",
        )
        with FileHandler() as fh:
            fh.create_directory(model_path)
        home_working_directory = os.getcwd()
        os.chdir(str(model_path))
        script_path = "../run_modeller_" + "".join(self.mutation) + ".py"
        command = f"{self.modeller_exec} {str(script_path)}"
        subprocess.run(
            command, shell=True, universal_newlines=True, check=True
        )
        os.chdir(home_working_directory)
        self.load_log_file()

    def load_log_file(self):
        """ Open the log file after the run of modeller.
        Look for the table with the DOPE scores.
        Save it as attribute (list).
        
        But if the mutation is not modellable, due to missing template
        converage, just save a warning string as attribute.
        """
        if self.mutation_is_modellable:
            with FileHandler() as fh:
                log_path = Path(self.isoform.out_path,"run_modeller_" + "".join(self.mutation) + ".log")
                logs=fh.read_file(log_path).splitlines()
                
                table_start_index=logs.index("Filename                          molpdf     DOPE score    GA341 score")
                table_end_index=table_start_index+logs[table_start_index:].index("")
                
                self.logged_scores=logs[table_start_index:table_end_index]
        else:
            self.logged_scores=["! The mutation was not modelled, due to missing templates for that region."]


from .database_parser import DatabaseParser
from .gene import Gene
from .input_parser import InputParser
from .analyzer import Analyzer

class IsoformRun:
    def __init__(self):
        self.parameters = []
        self.mutation_list = []
        self.input_parser = InputParser()
        self.genes = []

    def load_input(self, argv) -> None:
        """Load user input from the command line and parameters file
         and save them as attributes.

        :param argv: command line arguments
        """
        _, self.parameters = self.input_parser.load_input(argv)

    def load_mutation_list(self) -> None:
        """Parse the list of mutations and genes from the mutation_list file
        and save it as attribute.
        """
        self.mutation_list = self.input_parser.parse_input(
            self.parameters["INPUT_FILE"]
        )

    def create_genes(self) -> None:
        """Take the list of mutations of the genes saved
        in self.mutation_list, and for each gene, create a Gene instance.
        All Gene instances are saved in the
        self.gene list.
        """
        for i, gene_mutation_block in enumerate(self.mutation_list):
            self.genes.append(Gene(gene_mutation_block, self.parameters["OUTPUT_PATH"]))

    def create_isoforms(self) -> None:
        """Loads isoforms for each gene in the 'genes' attribute from
        the Uniprot database.
        """
        with DatabaseParser(self.parameters["DB_LOCATION"]) as db_parser:
            for gene in self.genes:
                gene.load_isoforms(db_parser)

    def run_blastp(self) -> None:
        """Start a blastp query for every loaded isoform. The proper
        function is a method of the class Isoform. This is used to
        split the whole in smaller steps.
        """
        for gene in self.genes:
            for isoform in gene.isoforms:
                isoform.blastp_search()

    def run_cobalt(self) -> None:
        """Start a cobalt run for every loaded isoform. The proper
        function is a method of the class Isoform. This is used to
        split the whole in smaller steps.
        """
        for gene in self.genes:
            for isoform in gene.isoforms:
                isoform.create_MSA(cobalt_home=self.parameters["COBALT_HOME"])

    def run_hmmsearch(self) -> None:
        """Start a hmmsearch for every loaded isoform. The proper
        function is a method of the class Isoform. This is used to
        split the whole in smaller steps.
        """
        for gene in self.genes:
            for isoform in gene.isoforms:
                isoform.HMMsearch(hmmer_home=self.parameters["HMMER_HOME"])

    def load_templates(self) -> None:
        for gene in self.genes:
            for isoform in gene.isoforms:
                isoform.load_templates(
                    int(self.parameters["PDB_TO_USE"]),
                    float(self.parameters["RESOLUTION"]),
                    self.parameters["COBALT_HOME"],
                )

    def select_isoforms(self) -> None:
        for gene in self.genes:
            gene.select_isoforms(
                float(self.parameters["W_STRUCT"]),
                float(self.parameters["W_MUT"]),
                float(self.parameters["SEQID"]),
                int(self.parameters["MODEL_CUTOFF"]),
                cobalt_home=self.parameters["COBALT_HOME"],
            )

    def start_modeller(self) -> None:
        """Run modeller only for the isoforms to model.
        The method write_modeller() of the isoform accepts the mutation
        as argument.
        """
        for gene in self.genes:
            gene.report_on_selected_isoforms()
            for isoform, mutation in gene.isoforms_to_model:
                isoform.run_modeller(
                    mutation, self.parameters["MODELLER_EXEC"],
                    int(self.parameters["MODEL_CUTOFF"]),
                    int(self.parameters["NUM_OF_MOD_WT"]),
                    int(self.parameters["NUM_OF_MOD_MUT"]),
                    self.parameters["COBALT_HOME"])

            

    def write_report(self):
        for gene in self.genes:
            gene.write_report()
            

class AnalysisRun:
    def __init__(self):
        self.parameters = []
        self.genes = []
        self.input_parser = InputParser()
        self.genes = []
        
    def load_input(self, argv) -> None:
        """Load user input from the command line and parameters file
         and save them as attributes.

        :param argv: command line arguments
        """
        _, self.parameters = self.input_parser.load_input(argv)
        self.pesto_home=self.parameters["PESTO_HOME"]
        self.output_path=self.parameters["OUTPUT_PATH"]

    def load_genes_from_mutation_list(self) -> None:
        """Parse the list of mutations and genes from the mutation_list file
        and save the genes as attribute.
        """
        self.genes = [m[0] for m in self.input_parser.parse_input(
            self.parameters["INPUT_FILE"])]
        
    def analysis(self) -> None:
        self.analyzer=Analyzer(self.pesto_home,self.output_path,
                                    self.genes)
        



from pathlib import Path
from typing import Union

from Bio.PDB import PDBIO, PDBList, PDBParser, Select, Selection, MMCIFParser
from Bio.PDB.Polypeptide import index_to_one, three_to_index

from .file_handler import FileHandler


class ChainSelection(Select):
    def __init__(self, chain_letters, standard_atoms=True):
        self.chain_letters = chain_letters.upper()
        self.standard_atoms = standard_atoms
        self.standard_residues = [
            "ALA",
            "ARG",
            "ASN",
            "ASP",
            "CYS",
            "GLU",
            "GLN",
            "GLY",
            "HIS",
            "ILE",
            "LEU",
            "LYS",
            "MET",
            "PHE",
            "PRO",
            "SER",
            "THR",
            "TRP",
            "TYR",
            "VAL",
        ]

        self.first_model = True  # see accet_model method

    def accept_model(self, model):
        # Accept only the first model
        # The attribute is initialized as True
        # and turns to False after the first run.

        if self.first_model:
            self.first_model = False
            return True
        else:
            return False

    def accept_residue(self, residue):
        # Accept only residues that are in the list of standard amino acids
        return residue.resname.strip() in self.standard_residues

    def accept_chain(self, chain):
        # Filter the chain
        return chain.id == self.chain_letters

    def accept_atom(self, atom):
        # Filter for standard atoms if requested
        if self.standard_atoms:
            return True
        return False


class PDB_manager:
    def __init__(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def downloadPDB(self, pdb: str, out_path: Union[str, Path]) -> Path:
        right_pdbname = f"{pdb}.pdb"
        wrong_pdbname = f"pdb{pdb}.ent"
        download_dir = ".pdbdownloads"

        filepath = Path(out_path, download_dir, right_pdbname)
        with FileHandler() as fh:
            if not fh.check_existence(Path(out_path, download_dir)):
                fh.create_directory(Path(out_path, download_dir))

            if fh.check_existence(filepath):
                return filepath
            else:
                pdbl = PDBList()
                filename = pdbl.retrieve_pdb_file(
                    pdb,
                    pdir=str(Path(out_path, download_dir)),
                    file_format="pdb",
                    overwrite=True,
                    obsolete=False
                )

                if fh.check_existence(filename):
                    fh.move_file(
                        Path(out_path, download_dir, wrong_pdbname), filepath
                    )
                    return filepath
                else:
                    RuntimeWarning(f"Could not download PDB {pdb}. Now trying to find the mmCIF format...")
                filepath = Path(out_path, download_dir, right_pdbname.replace("pdb","cif"))
                filename = pdbl.retrieve_pdb_file(
                    pdb,
                    pdir=str(Path(out_path, download_dir)),
                    overwrite=True,
                    obsolete=False
                )

                if fh.check_existence(filename):
                    fh.move_file(
                        filename, filepath
                    )
                    return filepath
                else:
                    FileNotFoundError(f"Could not download mmCIF either. Structure {pdb} not found.")


    def extract_clean_chain(
        self,
        input_pdb_path: Union[Path, str],
        output_pdb_path: Union[Path, str],
        chain_letter: str,
        resolution_cutoff: float,
    ):
        """Take an input path, save the standard atoms
        of chain 'chain_letter' in
        a filtered new PDB file, if resolution is better
        than the parameter 'resolution'.
        The exception is with the NMR structures that will
        be included in all cases.

        Args:
            input_pdb_path (Union[Path,str]): The original PDB file path
            to be filtered
            output_pdb_path (Union[Path,str]): The path of the output PDB
            chain_letter (str): Letter of the chain to extract.

        Returns:
            resolution (float or None): The resolution of the
            X-Ray or CryoEM structure
        """
        if str(input_pdb_path).endswith("pdb"):
            parser = PDBParser(QUIET=True)
        else:
            parser = MMCIFParser()
        structure = parser.get_structure("structure", str(input_pdb_path))
        if structure.header["resolution"]:
            if structure.header["resolution"] <= resolution_cutoff:
                with FileHandler() as fh:
                    io = PDBIO()
                    io.set_structure(structure)
                    if not fh.check_existence(output_pdb_path):
                        io.save(
                            str(output_pdb_path), ChainSelection(chain_letter)
                        )

                    # remove HETATM from saved file
                    saved_file = fh.read_file(output_pdb_path).splitlines()
                    saved_file = "\n".join(
                        [line for line in saved_file if "HETATM" not in line]
                    )
                    fh.write_file(output_pdb_path, saved_file)
                    return structure.header["resolution"]

        else:
            # The content of the following "if" cleans the NMR structures
            # But it's still missing an estimate of the reolution. Therefore
            # these structures will be cleaned, and included in all cases..
            if "nmr" in structure.header["structure_method"]:
                with FileHandler() as fh:
                    io = PDBIO()
                    io.set_structure(structure)
                    if not fh.check_existence(output_pdb_path):
                        io.save(
                            str(output_pdb_path), ChainSelection(chain_letter)
                        )
                print(f"NMR structure in path {output_pdb_path}")
                return 0

        print(
            f"The file {str(input_pdb_path)} was "
            + "exluded due to poor resolution."
        )
        return 9999

    def extract_fasta(
        self,
        pdb_name: str,
        pdb_path: Union[str, Path],
        output_fasta_path: Union[str, Path],
    ) -> None:
        """Load a PDB file and return the fasta sequence as a string.

        Args:
            pdb_name: Name of the PDB file and chain letter
            pdb_path (Union[str,Path]): Path to the PDB structure.
            output_fasta_path (Union[str,Path]): Path where to save the
            fasta sequence.
        """
        with FileHandler() as fh:
            parser = PDBParser(QUIET=True)
            structure = parser.get_structure("structure", str(pdb_path))
            io = PDBIO()
            io.set_structure(structure)

            residues = Selection.unfold_entities(structure, "R")
            resnames = [x.get_resname() for x in residues]
            sequence = "".join(
                [index_to_one(three_to_index(resname)) for resname in resnames]
            )
            content = ">" + pdb_name + "\n"
            content += sequence
            content += "\n"
            # This next check could be moved at the beginning
            # and if .fasta file already exists, it should
            # be loaded
            if not fh.check_existence:
                fh.write_file(output_fasta_path, content)
            return sequence



import argparse
from pathlib import Path
from typing import Dict, List


class InputParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_arguments()

    def add_arguments(self) -> None:
        """Add a list of expected arguments to the Parser instance.

        :return: None
        """
        super().add_argument(
            "-i",
            "--input_file",
            help="Path to the file genes and mutations",
            type=str,
            required=True,
        )

        super().add_argument(
            "-o",
            "--out_path",
            help="Path to the output folder",
            type=str,
            required=True,
        )

        super().add_argument(
            "-pf",
            "--par_file",
            help="Input the path to the parameters file",
            type=str,
            required=False,
        )
        
        super().add_argument(
            "-pst",
            "--pesto_home",
            help="path to the directory containing the\
                PeSto repo downloaded from GitHub",
            type=str,
            required=False,
        )
        
        super().add_argument(
            "-db",
            "--db_home",
            help="path to the DBs containing the\
                canonical and isoform sequences",
            type=str,
            required=False,
        )

        super().add_argument(
            "-cobalt",
            "--cobalt_home",
            help="path to COBALT bin folder",
            type=str,
            required=False,
        )

        super().add_argument(
            "-hmmer",
            "--hmmer_home",
            help="path to HMMER bin folder",
            type=str,
            required=False,
        )

        parameters_group = super().add_argument_group(
            "input", "manually provides the inputs"
        )
        parameters_group.add_argument(
            "-res",
            "--resolution",
            help="worst accepted resolution",
            type=float,
            default=4.50,
            required=False,
        )
        parameters_group.add_argument(
            "-seqid",
            "--sequence_identity",
            help="minimum sequence identity accepted",
            type=float,
            default=25,
            required=False,
        )
        parameters_group.add_argument(
            "-max_pdb",
            "--max_pdb_templates",
            help="maximum number of PDBs to use as templates",
            type=int,
            default=10,
            required=False,
        )

        parameters_group.add_argument(
            "-wt_mod",
            "--max_model_wt",
            help="maximum number of wild-type models to build",
            type=int,
            default=1,
            required=False,
        )

        parameters_group.add_argument(
            "-mut_mod",
            "--max_model_mut",
            help="maximum number of mutant models to build",
            type=int,
            default=1,
            required=False,
        )

        parameters_group.add_argument(
            "-rc",
            "--res_cutoff",
            help="maximum number of not crystalised residues\
                accepted before cutting a model",
            type=float,
            default=5.0,
            required=False,
        )

        parameters_group.add_argument(
            "-ws",
            "--w_struct",
            help="weight of the structural function",
            type=float,
            default=10.0,
            required=False,
        )

        parameters_group.add_argument(
            "-wm",
            "--w_mut",
            help="weight of the mutation function",
            type=float,
            default=10.0,
            required=False,
        )

        parameters_group.add_argument(
            "-mod",
            "--modeller_exec",
            help="path to the modeller executable",
            type=str,
            default="mod10.4",
            required=False,
        )

    def check_arguments(self, args: argparse.Namespace) -> None:
        """Verify that the necessary parameters have been provided

        :param args: passed arguments
        """
        if not args.par_file and (not args.db_home or not args.cobalt_home
                                  or not args.hmmer_home
                                  or not args.pesto_home):
            raise TypeError(
                "Specify a parameters file or insert the paths\
                    to the DBs, COBALT, PESTO and HMMER manually"
            )
        elif args.par_file and (
            args.db_home or args.cobalt_home or args.hmmer_home
        ):
            raise TypeError(
                "Either specify a parameters file or\
                    insert parameters manually"
            )

    def parse_input(self, mutation_file_path: argparse.Namespace) -> List:
        """Parse the list of mutations and genes from the mutation_list file.

        :param mutation_list: path to the file containing the list
        of mutations and genes
        :return: The list of gene and mutations
        """
        with Path(mutation_file_path).open() as my_file:
            content = my_file.read()
            
        
        #Remove useless new lines
        while content[-1]=="\n":
            content=content[:-1]
            
        while "\n\n\n" in content:
            content=content.replace('\n\n\n','\n\n')
        
        blocks = [block.splitlines() for block in content.split("\n\n")]
        return blocks

    def get_parameters(
        self,
        parameters_path: argparse.Namespace = "parameters.dat",
    ) -> Dict:
        """Collect the parameters from the parameters file if provided.

        :param parameters_path: Path to the parameters file,
        defaults to "parameters.dat".

        :return: Dict of keywords with the associated parameters
        """
        keywords = {
            "RESOLUTION": None,
            "SEQID": None,
            "PDB_TO_USE": None,
            "PESTO_HOME": None,
            "INPUT_FILE": None,
            "OUTPUT_PATH": None,
            "DB_LOCATION": None,
            "HMMER_HOME": None,
            "COBALT_HOME": None,
            "MODEL_CUTOFF": None,
            "NUM_OF_MOD_WT": None,
            "NUM_OF_MOD_MUT": None,
            "W_STRUCT": None,
            "W_MUT": None,
            "MODELLER_EXEC": None,
        }
        keys = list(keywords)
        if not Path(parameters_path).exists():

            error_message = f"Parameters file not found in {parameters_path},\
                please check the path provided."
            raise TypeError(error_message)

        with Path(parameters_path).open() as my_file:
            lines = my_file.readlines()

        for key in keys:
            for line in lines:
                if key in line:
                    value = line[line.find("=") + 1 :].strip()
                    keywords[key] = value
        return keywords

    def convertArgsToParameters(self, args: argparse.Namespace) -> Dict:
        """Make the parameters dictionary if they are passed as arguments

        :param args: passed arguments
        """
        
        return {
            "RESOLUTION": args.resolution,
            "SEQID": args.sequence_identity,
            "PDB_TO_USE": args.max_pdb_templates,
            "PESTO_HOME":args.pesto_home,
            "OUTPUT_PATH":args.out_path,
            "INPUT_FILE": args.input_file,
            "DB_LOCATION": args.db_home,
            "HMMER_HOME": args.hmmer_home,
            "COBALT_HOME": args.cobalt_home,
            "MODEL_CUTOFF": args.res_cutoff,
            "NUM_OF_MOD_WT": args.max_model_wt,
            "NUM_OF_MOD_MUT": args.max_model_mut,
            "W_STRUCT": args.w_struct,
            "W_MUT": args.w_mut,
            "MODELLER_EXEC": args.modeller_exec,
        }
        
    def merge_parameters(self,args,parameters):
        """Merges args and parameters, with a preference for parameters.
        
        The settings expressed in the command line will be added to the parameters dictionary.
        """
                
        args_name=[args.resolution,
              args.sequence_identity,
              args.max_pdb_templates,
              args.pesto_home,
              args.input_file,
              args.out_path,
              args.db_home,
              args.hmmer_home,
              args.cobalt_home,
              args.res_cutoff,
              args.max_model_wt,
              args.max_model_mut,
              args.w_struct,
              args.w_mut,
              args.modeller_exec
              ]
        
        dict_keys=['RESOLUTION',
                   'SEQID',
                   'PDB_TO_USE',
                   'PESTO_HOME',
                   'INPUT_FILE',
                   'OUTPUT_PATH',
                   'DB_LOCATION',
                   'HMMER_HOME',
                   'COBALT_HOME',
                   'MODEL_CUTOFF',
                   'NUM_OF_MOD_WT',
                   'NUM_OF_MOD_MUT',
                   'W_STRUCT',
                   'W_MUT',
                   'MODELLER_EXEC']
        
        
        for i, key in enumerate(parameters):
            if parameters[key] is None:
                try:
                    parameters[key]=args_name[i]
                except:
                    raise RuntimeError(f"The argument for {dict_keys[i]} was\
                    not found in the parameter file,\
                    and there was a error in reading it from the\
                    command line .")
                    
    def print_parameters(
        self, args: argparse.Namespace, parameters: Dict
    ) -> None:
        """Print the parameters provided

        :param parameters: Provided parameters
        """
        param = (
            f"\nResolution: {parameters['RESOLUTION']}\n"
            f"SEQ ID: {parameters['SEQID']}\n"
            f"WT MODELS TO PREPARE: {parameters['NUM_OF_MOD_WT']}\n"
            f"MUTANTS MODEL TO PREPARE: {parameters['NUM_OF_MOD_MUT']}\n"
            f"MAX PDBS AS TEMPLATES: {parameters['PDB_TO_USE']}\n"
            f"RESIDUES CUTOFF: {parameters['MODEL_CUTOFF']}\n"
            f"INPUT FILE: {args.input_file}\n"
            f"OUTPUT DIRECTORY: {args.out_path}\n"
            f"DATABASES DIRECTORY: {parameters['DB_LOCATION']}\n"
            f"COBALT: {parameters['COBALT_HOME']}\n"
            f"PESTO: {parameters['PESTO_HOME']}\n"
            f"HMMER: {parameters['HMMER_HOME']}\n"
            f"WEIGHT STRUCTURAL SCORE: {parameters['W_STRUCT']}\n"
            f"WEIGHT MUTATION SCORE: {parameters['W_MUT']}\n"
            f"MODELLER EXECUTABLE: {parameters['MODELLER_EXEC']}"
        )

        print(param)

    def load_input(self, argv) -> tuple:
        """Load user input from the command line and parameters file.

        :param argv: command line arguments
        """
        args, unparsed = self.parse_known_args(argv)
        self.check_arguments(args)
        if args.par_file:
            parameters = self.get_parameters(args.par_file)
        else:
            parameters = self.convertArgsToParameters(args)
        
        self.merge_parameters(args,parameters)
        self.print_parameters(args, parameters)
        return (args, parameters)



from pathlib import Path
import datetime
from .file_handler import FileHandler

class DatabaseParser:
    def __init__(self, db_location: str):
        self.db_location = db_location
        self.load_database()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass
    
    def _check_file_age(self,file_creation_time):
        """Given a database Path, loads the date of the creation
        of the file. If it exceeds n_weeks, it raises a warning.
        
        The parameter file_creation_time is obtained from FileHandler.get_date()"""

        current_time = datetime.datetime.now()
        creation_time = datetime.datetime.fromtimestamp(file_creation_time)
        age = current_time - creation_time
        
        if age > datetime.timedelta(weeks=24):
            print(f"\n\n---WARNING---\n",
                  "The file database is older than 24 weeks.\nIt should be updated.\n----------")
    
    def parse_database(self, database_path: Path) -> list[list[str]]:
        """
        Splits the sequences from a database file into a list, where each
        list element corresponds to a single sequence.
        Each sequence is represented as a list of strings, with each string
        being a line from the database that contains
        a part of the sequence. This allows for multi-line sequences in the
        database to be captured fully in each element.

        :param database_path: The Path pointing to the Uniprot database
        :return: A list of gene sequences, each split in multiple lines
        """
        with FileHandler() as fh:
            content = fh.read_file(database_path)
            file_creation_time=fh.get_date(database_path)
            self._check_file_age(file_creation_time)
        return [
            block.splitlines()
            for block in content.split(">")
            if "sp|" in block
        ]

    def load_database(self):
        """
        Load Uniprot databases as local attributes, using the
        parse_database() method
        """
        canonical_db_path = Path(self.db_location, "uniprot_sprot.fasta")
        print("Loading canonical database...")
        self.canonical_db = self.parse_database(canonical_db_path)
        print(f"Loaded {len(self.canonical_db)} elements.")

        print("Loading split variants database...")
        isoforms_db_path = Path(
            self.db_location, "uniprot_sprot_varsplic.fasta"
        )
        self.isoforms_db = self.parse_database(isoforms_db_path)
        print(f"Loaded {len(self.isoforms_db)} elements.")

    def get_canonical_isoforms(self, gene_name: str) -> list[list[str]]:
        """Use the gene name to retrieve the corresponding sequences
        from the canonical database.

        :param gene_name: Name of the gene
        :return: List of multi-line gene sequences
        """
        gene_name = f"GN={str(gene_name.upper())}"
        species = "OS=Homo sapiens"
        output = [
            seq
            for seq in self.canonical_db
            if gene_name in seq[0] and species in seq[0]
        ]
        if len(output) == 0:
            print(
                f"No sequences found for gene {gene_name} in\
the canonical isoforms database."
            )
        return output

    def get_noncanonical_isoforms(self, gene_name: str) -> list[list[str]]:
        """Use the gene name to retrieve the corresponding sequences from
        the split variants database.

        :param gene_name: Name of the gene
        :return: List of multi-line gene sequences
        """
        gene_name = f"GN={str(gene_name.upper())}"
        species = "OS=Homo sapiens"
        output = [
            seq
            for seq in self.isoforms_db
            if gene_name in seq[0] and species in seq[0]
        ]
        if len(output) == 0:
            print(
                f"No sequences found for gene {gene_name} in \
the split variants isoforms database."
            )
        return output



import os
import shutil
from pathlib import Path
from typing import Union

class FileHandler:
    def __init__(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def remove_file(self, file_path: Union[str, Path]):
        """Removes a file at the specified path."""
        file_path = str(file_path)
        try:
            os.remove(file_path)
            print(f"File {file_path} removed successfully.")
        except FileNotFoundError:
            print(f"The file {file_path} does not exist.")
        except Exception as e:
            print(f"Error removing file {file_path}: {e}")

    def create_directory(self, dir_path: Union[str, Path]):
        """Creates a new directory at the specified path.
        It creates all the directory tree if it does not exist.
        """
        try:
            os.makedirs(str(dir_path), exist_ok=True)
            print(f"Directory {str(dir_path)} created successfully.")
        except Exception as e:
            print(f"Error creating directory {dir_path}: {e}")

    def move_file(self, src: Union[str, Path], dest: Union[str, Path]):
        """Moves a file from src to dest."""
        src = str(src)
        dest = str(dest)
        try:
            shutil.move(src, dest)
            print(f"File moved from {src} to {dest}.")
        except FileNotFoundError:
            print(f"The file {src} does not exist.")
        except Exception as e:
            print(f"Error moving file from {src} to {dest}: {e}")

    def copy_file(self, src: Union[str, Path], dest: Union[str, Path]):
        """Copies a file from src to dest."""
        src = str(src)
        dest = str(dest)
        try:
            shutil.copy(src, dest)
            print(f"File copied from {src} to {dest}.")
        except FileNotFoundError:
            print(f"The file {src} does not exist.")
        except Exception as e:
            print(f"Error copying file from {src} to {dest}: {e}")

    def write_file(self, file_path: Union[str, Path], content: str):
        """Writes content to a file, removing it first if it exists."""

        # Check if the file exists and remove it
        # using the class's remove_file method
        file_path = str(file_path)
        if os.path.exists(file_path):
            self.remove_file(file_path)

        # Now, write the new content to the file
        try:
            with open(file_path, "w") as file:
                file.write(content)
            print(f"Content written to {file_path} successfully.")
        except Exception as e:
            print(f"Error writing to file {file_path}: {e}")

    def check_existence(self, path: Union[str, Path]) -> bool:
        """
        Checks whether a given file or directory exists.

        Args:
            path (Union[str, Path]): The path of the file or directory
            to check.

        Returns:
            bool: True if the file or directory exists, False otherwise.
        """
        # Convert Path to string if it's not already a string
        path_str = str(path)
        return os.path.exists(path_str)

    def read_file(self, path: Union[str, Path]) -> str:
        """
        Read and return the content of a file after checking its existence.

        Args:
            path (Union[str, Path]): The path of the file or directory
            to check.

        Returns:
            str: The file content.
        """
        if not self.check_existence(path):
            raise (FileNotFoundError("The specified file was not found."))

        else:
            path = Path(path)
            with path.open("r") as file:
                content = file.read()
                return content
        
    def get_date(self,path: Union[str,Path]):
        """ Given a file, returns the creation a last-modified
            time as a tuple in "seconds since the epoch"

        Args:
            path (Union[str,Path]): The path of the file
        """
        # Both the variables would contain time
        # elapsed since EPOCH in float
        ti_c = os.path.getctime(path)
        ti_m = os.path.getmtime(path)

        return ti_c


import os
import sys
import numpy as np
import torch as pt 
from tqdm import tqdm
from glob import glob
from pathlib import Path
from Bio.PDB.SASA import ShrakeRupley
from Bio.PDB import PDBParser
from .file_handler import FileHandler


class Analyzer:
    # pylint: disable=import-error
    def __init__(self,pesto_path,output_path,gene_list):
        self.output_path=output_path
        self.gene_list=gene_list
        model_path=os.path.join(pesto_path,"model/save/i_v4_1_2021-09-07_11-21/")
        self.model_filepath=os.path.join(model_path,"model_ckpt.pt")
        if pesto_path not in sys.path:
            sys.path.insert(0, pesto_path)
        if model_path not in sys.path:
            sys.path.insert(0, model_path)
            
        from config import config_model # type: ignore
        from model import Model # type: ignore
        
        self.device=pt.device("cpu")
        self.model = Model(config_model)
        self.model.load_state_dict(pt.load(self.model_filepath, map_location=pt.device("cpu")))
        self.model = self.model.eval().to(self.device)
        
        self._findPDBfiles()
        self.runPestoAnalysis()
        self.runSasaAnalysis()
    

    def _findPDBfiles(self):
        
        from src.dataset import StructuresDataset # type: ignore
        self.pdb_filepaths=[]
        for gene in self.gene_list:
            self.pdb_filepaths+=glob(str(Path(self.output_path,gene,"*","*model","*pdb")))
        
        self.pdb_filepaths=[f for f in self.pdb_filepaths if "pesto" not in f]
        self.dataset=StructuresDataset(self.pdb_filepaths, with_preprocessing=True) # type: ignore
    
    def runPestoAnalysis(self):
        from src.structure import concatenate_chains # type: ignore
        from src.data_encoding import encode_structure, encode_features, extract_topology # type: ignore
        from src.dataset import collate_batch_features # type: ignore
        from src.structure import encode_bfactor,split_by_chain # type: ignore
        from src.structure_io import save_pdb # type: ignore
        
        #module 'numpy' has no attribute 'object'. The aliases was originally deprecated in NumPy 1.20
        #This fixes compatibility with the Numpy versio PeSto was programmed for.
        setattr(np,"object",object) 
        
        results=["Protein", "DNA-RNA", "Lipid", "Ligand", "Ion"]
        with pt.no_grad():
            for subunits, filepath in tqdm(self.dataset):
                # concatenate all chains together
                structure = concatenate_chains(subunits)

                # encode structure and features
                X, M = encode_structure(structure) 
                #q = pt.cat(encode_features(structure), dim=1)
                q = encode_features(structure)[0] 

                # extract topology
                ids_topk, _, _, _, _ = extract_topology(X, 64)

                # pack data and setup sink (IMPORTANT)
                X, ids_topk, q, M = collate_batch_features([[X, ids_topk, q, M]]) 

                # run model
                z = self.model(X.to(self.device), ids_topk.to(self.device), q.to(self.device), M.float().to(self.device))

                # for all predictions
                for i in range(z.shape[1]):
                    # prediction
                    p = pt.sigmoid(z[:,i])

                    # encode result
                    structure = encode_bfactor(structure, p.cpu().numpy())

                    # save results
                    output_filepath = filepath[:-4]+'_pesto_{}.pdb'.format(results[i])
                    save_pdb(split_by_chain(structure), output_filepath)
    
    def runSasaAnalysis(self) -> None:
        p = PDBParser(QUIET=1)
        with FileHandler() as fh:
            for pdb in self.pdb_filepaths:
                
                struct_sasa_path=Path(pdb.replace(".pdb",".struct.sasa"))
                residue_sasa_path=Path(pdb.replace(".pdb",".residue.sasa.csv"))

                struct = p.get_structure("Protein", pdb)
                sr = ShrakeRupley()
                sr.compute(struct, level="S")
                sr.compute(struct, level="R")
                sr.compute(struct, level="A")

                struct_sasa_string=f"Strcuture SASA: {struct.sasa}"
                fh.write_file(struct_sasa_path,struct_sasa_string)
                
                residue_sasa=[]
                backbone_sasa=[]
                sidechain_sasa=[]
                
                for res in struct.get_residues():
                    residue_sasa.append(res.sasa)
                    bb_atoms=[a for a in res.get_atoms() if a.name in ['C','N','CA','O']]
                    sc_atoms=[a for a in res.get_atoms() if a.name not in ['C','N','CA','O']]
                    backbone_sasa.append(sum([a.sasa for a in bb_atoms]))
                    sidechain_sasa.append(sum([a.sasa for a in sc_atoms]))
                    
                residue_sasa_string="Residue sasa, Backbone, Sidechain\n"
                
                for i,r in enumerate(residue_sasa):
                    residue_sasa_string+=f"{r},{backbone_sasa[i]},{sidechain_sasa[i]}\n"
                
                fh.write_file(residue_sasa_path,residue_sasa_string)


from pathlib import Path
from typing import Union

from .file_handler import FileHandler
from .PDB_manager import PDB_manager


class Template:
    def __init__(
        self,
        pdb_name: str,
        out_path: Union[str, Path],
        gene_name: str,
        isoform_name: str,
        resolution_cutoff,
    ):
        """Assign PDB attributes to this object, such as name,
        chain, gene_name, etc. Within the out_path directory of
        the isoform, create the templates directory, download the
        pdb file, clean it, and extract resolution and fasta sequence.

        Args:
            pdb_name (str): The pdb name followed by the
            chain letter (e.g., "6J8G_A").

            out_path (Union[str,Path]): Path to the isoform directory.
            In the way the code is written, you also need access to the
            parent of the parent.
            gene_name (str): name of the gene.
            isoform_name (str): name of the isoform.
        """

        self.usable = True  # or False after checks
        self.pdb_name = pdb_name[:4]
        self.pdb_chain = pdb_name[-1].upper()
        self.gene_name = gene_name
        self.out_path = Path(out_path)
        self.isoform_name = isoform_name
        self.templates_directory = Path(self.out_path, "templates")
        self.pdb_filename = Path(
            self.templates_directory, self.pdb_name + ".pdb"
        )
        self.mmCIF_filename = Path(
            self.templates_directory, self.pdb_name + ".cif"
        )
        self.clean_pdb_filename = Path(
            self.templates_directory, pdb_name + "_clean.pdb"
        )
        self.clean_fasta_file = Path(
            self.templates_directory, pdb_name + ".fasta"
        )
        self.resolution_cutoff = resolution_cutoff
        self.resolution = 9999  # Overwritten with Xray and CryoEM resolution
        self.sequence = ""
        self.aligned_sequence = ""
        self.clean_aligned_sequence = ""
        self.sequence_identity = 0

        self.get_pdb_file()
        self.get_clean_pdb_chain()
        if self.usable:
            self.get_fasta()

    def get_pdb_file(self) -> None:
        """Create the template directory for the PDB files if does not exist.
        Download the PDB files that could be used as templates.

        Args:
            templates_list (str): list of pdbs that could be used for modelling
        """
        with FileHandler() as fh:
            if not fh.check_existence(self.templates_directory):
                fh.create_directory(self.templates_directory)

            with PDB_manager() as pm:
                if (not fh.check_existence(self.pdb_filename)) and (not fh.check_existence(self.mmCIF_filename)):
                    file = pm.downloadPDB(
                        self.pdb_name, self.out_path.parent.parent
                    )
                    fh.copy_file(file, self.templates_directory)

    def get_clean_pdb_chain(self) -> None:
        """Take the original PDB file in the template directory,
        extract the standard atoms from the single chain of interest,
        save it as a new file.
        """
        with PDB_manager() as pm:
            with FileHandler() as fh:
                if fh.check_existence(self.pdb_filename):
                    file=self.pdb_filename
                else:
                    file=self.mmCIF_filename
            self.resolution = pm.extract_clean_chain(
                file,
                self.clean_pdb_filename,
                self.pdb_chain,
                self.resolution_cutoff,
            )
            if self.resolution > self.resolution_cutoff:
                self.usable = False

    def get_fasta(self) -> None:
        """Load the clean PDB file of the chain
        of interest and extract the fasta
        sequence. Save it as an attribute.
        """
        with PDB_manager() as pm:
            self.sequence = pm.extract_fasta(
                self.pdb_name, self.clean_pdb_filename, self.clean_fasta_file
            )

    def add_aligned_sequence(self, aligned_sequence: str) -> None:
        """Method to save as attribute the aligned sequece.
        Invoked by parent Isoform object, after cobal alignment.
        """
        self.aligned_sequence = aligned_sequence
        
    def add_clean_aligned_sequence(self, clean_aligned_sequence: str) -> None:
        """Method to save as attribute the aligned sequece with chain
        breaks.
        """
        self.clean_aligned_sequence = clean_aligned_sequence

    def calculate_sequence_identity(self, reference_sequence: str) -> None:

        # Check if own sequence is loaded
        if len(self.clean_aligned_sequence) == 0:
            raise (
                RuntimeError(
                    f"Aligned sequence of template {self.pdb_name}"
                    " has length 0. Calculate structural score first."
                )
            )


        # If sequences are aligned, they should have the same length
        if len(self.clean_aligned_sequence) != len(reference_sequence):
            raise (
                RuntimeError(
                    "Length of the aligned template and isoform sequence "
                    "do not match.\n"
                    + self.aligned_sequence
                    + "\n"
                    + reference_sequence
                )
            )

        
        reference_sequence_length = len(reference_sequence.replace("-", ""))
        matching_residues = 0
        for residue_index, residue in enumerate(reference_sequence):
            if residue != "-" and residue !="/":
                if residue == self.clean_aligned_sequence[residue_index]:
                    matching_residues += 1

        self.sequence_identity = (
            matching_residues / reference_sequence_length * 100
        )



import subprocess
from pathlib import Path
from typing import Union

from Bio import SeqIO
from Bio.Blast import NCBIWWW as blastq
from Bio.Blast import NCBIXML as blastparser

from .cobalt_wrapper import Cobalt
from .file_handler import FileHandler
from .modeller_manager import Modeller_manager
from .template import Template


class Isoform:
    def __init__(
        self,
        gene_name: str,
        sequence: list[str],
        isoform_index: int,
        out_path: Union[str, Path],
        mutations: list,
    ):
        self.gene_name = gene_name
        self.isoform_index = isoform_index
        self.isoform_name = "isoform" + str(self.isoform_index)
        self.mutations = mutations
        self.mutations_not_in_structure=[]
        self.first_line = sequence[0]
        self.sequence = sequence[1:]
        self.out_path = Path(out_path, self.isoform_name)

        # just for initialization
        self.mutation_score = 0
        self.structural_score = 0
        self.selection_score = 0
        self.templates = []
        self.modellable = True  # False if all templates get excluded
        self.aligned_sequence = ""
        self.clean_aligned_sequence=""
        
        self.modeller_run=None
        
        self.create_directory()
        self.save_fasta_sequence()

    def create_directory(self) -> None:
        """Create an empty subdirectory with the isoform index
        at the output path specified by the user, and within
        the gene folder."""

        with FileHandler() as fh:
            fh.create_directory(Path(self.out_path))

    def save_fasta_sequence(self) -> None:

        text_output = "\n".join([">" + self.first_line] + self.sequence)
        with FileHandler() as fh:
            fh.write_file(
                Path(self.out_path, self.isoform_name + ".fasta"), text_output
            )

    def _check_hmmer_errors(self,possible_templates_string):
        if "Error" in possible_templates_string:
            print("The hmmsearch for a template resulted in a server error.")
            print("Monviso will quit.")
            quit()

    def blastp_search(self) -> None:
        """Use the isoform.fasta file saved in the directory
        to start a Blastp search. If the file already exists,
        nothing is done.
        """
        print(
            f"Looking for homologues of {self.gene_name} {self.isoform_name}"
        )
        file_path = Path(self.out_path, self.isoform_name + ".fasta")
        out_path = Path(self.out_path, self.isoform_name + "_hits.fasta")

        with FileHandler() as fh:
            if fh.check_existence(out_path):
                print(
                    "Blastp search output file " "already present in folder."
                )

            else:
                fasta_file = SeqIO.read(file_path, "fasta")

                results = blastq.qblast(
                    "blastp",
                    "swissprot",
                    fasta_file.seq,
                    alignments=500,
                    word_size=6,
                )
                blastRecord = blastparser.read(results)
                text_output = (
                    ">"
                    + fasta_file.id
                    + "\n"
                    + str(fasta_file.seq).strip()
                    + "\n"
                )
                for alignment in blastRecord.alignments:
                    for hsp in alignment.hsps:
                        text_output += f">{alignment.hit_id}\n"
                        text_output += str(hsp.sbjct).replace("-", "") + "\n\n"
                fh.write_file(
                    str(file_path).replace(".fasta", "_hits.fasta"),
                    text_output,
                )
        print("Done")

    def create_MSA(self, cobalt_home: Union[str, Path]) -> None:
        """Take the blastp results saved in the _hits.fasta file and use them
        as query for cobalt, if the MSA are not already present in the isoform
        folder.

        Args:
            cobalt_home (Union[str,Path]): Home of the Cobalt program,
            where executables are stored.
        """
        hits_path = Path(self.out_path, self.isoform_name + "_hits.fasta")
        aligned_path = Path(self.out_path, "aligned.fasta")
        with FileHandler() as fh:
            if fh.check_existence(aligned_path):
                print("Cobalt output file already present in folder.")
            else:
                with Cobalt() as cobalt:
                    cobalt.run(hits_path, aligned_path, cobalt_home)

    def buildHMM(self, hmmer_home: Union[str, Path]) -> Union[str, Path]:
        """Take the aligned cobalt output from the aligned.fast file
        and use it as query for  hmmbuild.

        Args:
            hmmer_home (Union[str,Path]): Home of the HMMer program,
            where executables are stored.

        Returns:
            output_path (Union[str,Path]): Path to .hmm file
        """
        print(f"Building HMM for gene {self.gene_name} {self.isoform_name}")
        output_path = Path(self.out_path, self.isoform_name + ".hmm")
        aligned_path = Path(self.out_path, "aligned.fasta")
        with FileHandler() as fh:
            if fh.check_existence(output_path):
                print("HMMsearch output file already present in folder.")
                return output_path
            else:
                command = f"{str(Path(hmmer_home,'hmmbuild'))} {output_path} {aligned_path}"
                subprocess.run(
                    command, shell=True, universal_newlines=True, check=True
                )
                return output_path
        print("Done")

    def HMMsearch(self, hmmer_home: Union[str, Path]) -> None:
        """Take the aligned cobalt output from the aligned.fasta
        file, build .hmm file,and use it as query for a hmmsearch.

        Args:
            hmmer_home (Union[str,Path]): Home of the HMMer program,
            where executables are stored.
        """
        with FileHandler() as fh:
            hmm_path = self.buildHMM(hmmer_home=hmmer_home)
            if not fh.check_existence(hmm_path):
                raise (FileNotFoundError(".hmm file not found."))

            print(
                "Looking for templates for "
                f"{self.gene_name} {self.isoform_name}"
            )
            templates_path = Path(self.out_path, "possible_templates.xml")

            if not fh.check_existence(templates_path):
                command = f"curl -L -H 'Expect:' -H \
                    'Accept:text/xml' -F seqdb=pdb -F\
                    seq='<{str(hmm_path)}' \
                    https://www.ebi.ac.uk/Tools/hmmer/search/hmmsearch"
                try:

                    result = subprocess.run(
                        command,
                        shell=True,
                        universal_newlines=True,
                        capture_output=True,
                        check=True,
                    )
                    output = result.stdout  # Captured output as a string
                    self._check_hmmer_errors(output)
                    fh.write_file(templates_path, output)
                except subprocess.CalledProcessError as e:
                    print(f"Error executing curl command: {e}")
            else:
                print(
                    f"Templates file for {self.gene_name} "
                    f"{self.isoform_name} esists. "
                    "Skipping hmmsearch."
                )

    def _extract_pdb_names(self, max_pdb: int) -> list:
        """Extact PDB names from the hmmsearch file found in
        the isoform directory, with the name possible
        templates.xml.

        Args:
            max_pdb: The maximum number of PDB templates
            to use to model the isoform.

        Returns:
            pdb_list (list(str)): A list of the PDBids and
            chains that can be used for the modeling.
        """
        with FileHandler() as fh:
            pdb_list_path = Path(self.out_path, "all_pdbs.dat")
            top_templates_path = Path(self.out_path, "top_templates.dat")
            templates_path = Path(self.out_path, "possible_templates.xml")
            if not (
                fh.check_existence(pdb_list_path)
                and fh.check_existence(top_templates_path)
            ):
                templates_content = fh.read_file(templates_path).splitlines()
                pdb_list = [
                    line[16:22]
                    for line in templates_content
                    if "hits name" in line
                ]
                templates_list = pdb_list[:max_pdb]
                fh.write_file(top_templates_path, "\n".join(templates_list))
                fh.write_file(pdb_list_path, "\n".join(pdb_list))
            templates_list = fh.read_file(top_templates_path).splitlines()
            return templates_list

    def _template_alignment(self, cobalt_home: Union[str, Path],redo=False) -> None:
        """Create a unique file with the templates sequences and
        run a cobalt MSA.
        Args:
            cobalt_home (Union[str,Path]): Home of the Cobalt program,
            where executables are stored.
            redo: True if part of the scoring cycle. Alignment can be re-written
        """
        content = ">" + self.gene_name + " " + self.isoform_name + "\n"
        content += "".join(self.sequence) + "\n"
        for template in self.templates:
            content += (
                ">" + template.pdb_name + "_" + template.pdb_chain + "\n"
            )
            content += template.sequence + "\n"

        with FileHandler() as fh:
            templates_path = Path(self.out_path, "templates_sequences.fasta")
            fh.write_file(templates_path, content)
            aligned_path = Path(self.out_path, "templates_aligned.fasta")
            if fh.check_existence(aligned_path) and not redo:
                print(
                    "Cobalt output file for templates "
                    "already present in folder."
                )
            else:
                with Cobalt() as cobalt:
                    cobalt.run(templates_path, aligned_path, cobalt_home)
                    print(
                        f"Cobalt alignment for {self.gene_name}"
                        f" {self.isoform_name} templates done."
                    )

    def load_templates(
        self,
        max_pdb: int,
        resolution_cutoff: float,
        cobalt_home: Union[str, Path],
    ):
        """Load a list of N (=max_pdb) PDB ids to use as templates.
        Create a Template object for each of the pdb files. Append
        the new object to self.templates.

        Args:
            max_pdb (int): Maximum number of PDB templates to use.
            resolution_cutoff (float): Accept CryoEM and Xray structures within
            this resolution cut-off.
            cobalt_home (Union[str,Path]): Home of the Cobalt program,
            where executables are stored.
        """
        templates_list = self._extract_pdb_names(max_pdb)
        for pdb in templates_list:
            template = Template(
                pdb,
                self.out_path,
                self.gene_name,
                self.isoform_name,
                resolution_cutoff,
            )
            if template.usable:
                self.templates.append(template)

        self._template_alignment(cobalt_home)
    
    def _add_chain_breaks(self, sequences: list, model_cutoff:int) -> list:
        """For alignments inw which there is no coverage for model_cutoff+
        residues, the section without coverage is replaced by
        chain breaks.

        Args:
            sequences (list): A list of [name, sequence] lists,
            one for each sequence to be written in the modeller
            alignment input file.

        Returns:
            list: Returns the same list, with chain breaks.
        """

        # Separate protein names and aligned sequences in two objects
        names = [object[0] for object in sequences]
        aligned_seq = [object[1] for object in sequences]

        # Calculate worst case number of residues without coverage
        max_non_covered = max([seq.count("-") for seq in aligned_seq])

        # Search for non-covered subsequences with lenght between
        # max_non_covered and model_cutoff
        while max_non_covered >= model_cutoff:

            # check if "-" repeated max_non_covered times
            # is present in all the aligned structures (except target seqence)
            check_presence = [
                "-" * max_non_covered in seq for seq in aligned_seq[1:]
            ]
            if sum(check_presence) == (len(aligned_seq) - 1):

                # Find position of the sequence of "-"'s in all
                # aligned sequences for each position it will
                # check if the coverage is missing in all the
                # sequences
                positions = [
                    seq.find("-" * max_non_covered) for seq in aligned_seq[1:]
                ]
                for pos in positions:
                    check_position = [
                        seq[pos : pos + max_non_covered]
                        == "-" * max_non_covered
                        for seq in aligned_seq[1:]
                    ]
                    if sum(check_position) == (len(aligned_seq) - 1):
                        # If that is the case, replace that section in
                        # all sequences (+ target sequences) with a chain
                        # break ("/").
                        for i, seq in enumerate(aligned_seq):
                            aligned_seq[i] = (
                                seq[:pos] + "/" + seq[pos + max_non_covered :]
                            )

                        max_non_covered += (
                            1  # This repeats the check for the current value
                        )
            max_non_covered -= 1

        sequences = [[names[i], aligned_seq[i]] for i in range(len(names))]
        return sequences


    def calculate_mutation_score(self, mappable_mutations: list) -> None:
        """Calculate and save as local attribute the score
        of the mutation function. Calculated as: mutations
        that can me mapped on the isoform divided by the number
        of mutations that can be mapped on at least one
        isoform of the gene.

        Args:
            mappable_mutations (list): The list of mutations that can be mapped
                                       on at least one isoform of the gene.
        """

        score = len(self.mutations) / len(mappable_mutations)
        self.mutation_score = score

    def calculate_structural_score(self, model_cutoff: int) -> None:
        """Load the sequence of the templates for the aligned file.
        Add the aligned sequence as an attribute of the template.
        Calculate how many residues of the target sequence are covered
        by at least one template.
        """
        aligned_templates_path = Path(self.out_path, "templates_aligned.fasta")
        with FileHandler() as fh:
            alignment = fh.read_file(aligned_templates_path).split(">")

            total_number_residues = 0  # to increase when reading alignment
            modellable_residues = 0  # to be increase when reaading alignment

            self.aligned_sequence = "".join(alignment[1].splitlines()[1:])

            templates_alignment = []
            for alignment_index, aligned_template in enumerate(alignment[2:]):
                template_sequence = "".join(aligned_template.splitlines()[1:])
                templates_alignment.append(template_sequence)
                self.templates[alignment_index].add_aligned_sequence(
                    template_sequence
                )
                
            # This part should be moved, but since the alignment is
            # loaded, here we calculate the sequence with chain breaks.
            # I.e., gaps longer than the value model_cutoff are removed.
            # The clean sequence is saved as Template's attribute to 
            # Calculate sequence identity of the modellable regions.
            
            raw_sequence=[self.aligned_sequence]+templates_alignment
            #Make it 2D for the way the next method is written
            raw_sequence=[[None, seq] for seq in raw_sequence]

            clean_sequence=self._add_chain_breaks(raw_sequence,model_cutoff)
            self.clean_aligned_sequence=clean_sequence[0][1]
            
            for template_id,template in enumerate(self.templates):
                template.add_clean_aligned_sequence(clean_sequence[template_id+1][1])
            
            ## End of the section for chain breaks.   
                
            for residue_index, residue in enumerate(self.aligned_sequence):
                if residue != "-":
                    total_number_residues += 1
                    if (
                        sum(
                            [
                                template[residue_index] != "-"
                                for template in templates_alignment
                            ]
                        )
                        > 0
                    ):
                        modellable_residues += 1

            score = modellable_residues / total_number_residues
            self.structural_score = score

    def filter_templates_by_sequence_identity(
        self, sequence_identity_cutoff: float
    ) -> None:
        """Take the list of templates and exclude
        the ones with a sequence identity lower or
         equal to the cutoff.

        Args:
            sequence_identity_cutoff (float): cut-off value
        """
                
        filtered_list = []
        for template in self.templates:
            template.calculate_sequence_identity(self.clean_aligned_sequence)
            if template.sequence_identity > sequence_identity_cutoff:
                filtered_list.append(template)
            else:
                print(
                    f"{template.sequence_identity} <"
                    f" {sequence_identity_cutoff}"
                )
                print(
                    f"Template {template.pdb_name} excluded"
                    " for low sequence identity"
                )

        if len(filtered_list) == 0:
            print(
                "Resolution and sequence identity cutoff"
                " let to an empty list of"
                f" templates for {self.gene_name} {self.isoform_name}"
            )
            self.modellable = False
        self.templates = filtered_list

    def calculate_selection_score(
        self, w1: float = 10.0, w2: float = 10.0
    ) -> None:
        """Calculate and save the selection score as attribute.
        The score is calculated with:
        score= w1*structural_score + w2*mutation_score

        Args:
            w1 (float): weight of the structural score
            w2 (float): weight of the mutation score
        """

        self.selection_score = (
            w1 * self.structural_score + w2 * self.mutation_score
        )

                    
    def calculate_score(self,mappable_mutations,model_cutoff,sequence_identity_cutoff,w1,w2,cobalt_home):
        """Starts a cycle to calculate the mutation, structural and selection scores.
        The process is repeated in cycle because, after calculating the sequence identity,
        the number of sequences, and thus the alignments might change. The cycle is repeated
        untill these values stabilize on a fixed value. After every calculation of the 
        sequence identy, a new alignment must be created with the filtered sequences.
        *indirectly: while the list of templates changes, the cycle repeats
        
        Args:
        - mappable mutations: A list of mutations that can be mapped on the sequence. 
        - model_cutoff: The maximum number of gaps in a region without coverage. Larger
        gaps are replaced by chain breaks.
        - sequence_identity_cutoff: Sequences with a lower sequence identity will be skipped.
        - w1 (float): weight of the structural score
        - w2 (float): weight of the mutation score
        """
        last_template_list=[]
        
        while last_template_list!=[t.pdb_name+"_"+t.pdb_chain for t in self.templates]:
            last_template_list=[t.pdb_name+"_"+t.pdb_chain for t in self.templates]
            #TO DO
            #Use of next function in self.load_templates should be avoided
            self._template_alignment(cobalt_home,redo=True)
            self.calculate_mutation_score(mappable_mutations)
            self.calculate_structural_score(model_cutoff)
            self.filter_templates_by_sequence_identity(
                    sequence_identity_cutoff
                )
            self.calculate_selection_score(w1, w2)

    def run_modeller(self, mutation, modeller_exec: str,model_cutoff:int,number_of_wt:int,number_of_mut:int,cobalt_home:str):
        """Create the alignment and start the Modeller run."""
        self.modeller_run=Modeller_manager(
            self, mutation, modeller_exec, model_cutoff,number_of_wt,number_of_mut)

        
        self.modeller_run.write()
        self.modeller_run.run()



"""Python setup.py for monviso_reloaded package"""
import io
import os
from setuptools import find_packages, setup


def read(*paths, **kwargs):
    """Read the contents of a text file safely.
    >>> read("monviso_reloaded", "VERSION")
    '0.1.0'
    >>> read("README.md")
    ...
    """

    content = ""
    with io.open(
        os.path.join(os.path.dirname(__file__), *paths),
        encoding=kwargs.get("encoding", "utf8"),
    ) as open_file:
        content = open_file.read().strip()
    return content


def read_requirements(path):
    return [
        line.strip()
        for line in read(path).split("\n")
        if not line.startswith(('"', "#", "-", "git+"))
    ]


setup(
    name="monviso",
    python_requires='>=3.9',
    version=read("monviso_reloaded", "VERSION"),
    description="MoNvIso is a comprehensive software tool designed for the analysis and modeling of protein isoforms. It automates the process of identifying canonical and additional isoforms, assessing their modeling propensity, mapping mutations accurately, and building structural models of proteins.",
    url="https://github.com/alisamalb/monviso_reloaded",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="S. Albani",
    packages=find_packages(exclude=["tests", ".github"]),
    install_requires=read_requirements("requirements.txt"),
    entry_points={
        "console_scripts": ["monviso = monviso_reloaded.__main__:main"]
    },
    extras_require={"test": read_requirements("requirements-test.txt")},
)



import sys

from monviso_reloaded.base import IsoformRun
from monviso_reloaded.base import AnalysisRun

def isoform(argv):
    run = IsoformRun()
    run.load_input(argv[1:])
    run.load_mutation_list()
    run.create_genes()
    run.create_isoforms()
    run.run_blastp()
    run.run_cobalt()
    run.run_hmmsearch()
    run.load_templates()
    run.select_isoforms()
    run.start_modeller()
    run.write_report()

def analysis(argv):
    run = AnalysisRun()
    run.load_input(argv[1:])
    run.load_genes_from_mutation_list()
    run.analysis()

def unrecognized_command():
    print("The command was not recognized. The available commands are:")
    print(" - monviso isoform")
    print(" - monviso analysis")
    print("append \"--help\" at the end to receive help on the tool.")
    
def main(argv=None):  # pragma no cover
    """
    Main function

    :param argv: argv

    :return: None
    """
    # arguments and parameters
    if argv is None:
        argv = sys.argv[1:]
    
    if len(argv)>0:
        if argv[0]=="isoform":
            isoform(argv)

        elif argv[0]=="analysis":
            analysis(argv)
        else:
            unrecognized_command()
    else:
        unrecognized_command()

def init() -> None:
    if __name__ == "__main__":
        sys.exit(main())


init()



import subprocess
from pathlib import Path
from typing import Union


class Cobalt:
    def __init__(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def run(
        self,
        hits_file: Union[str, Path],
        aligned_file: Union[str, Path],
        cobalt_home: str,
    ) -> bool:
        """_summary_

        Args:
            hits_file (Union[str,Path]): Path to the blastp output
            aligned_file (Union[str,Path]): Path where to direct the
            Cobalt output.

            cobalt_home (str): cobalt home, containing the executables

        Returns:
            bool: True if successful, else False
        """
        hits_file = Path(hits_file)
        aligned_file = Path(aligned_file)

        print("Doing MSA with COBALT.")
        command = [
            f"{cobalt_home}/cobalt",
            "-i",
            str(hits_file),
            "-outfmt",
            "mfasta",
            "-end_gapopen",
            "5",
            "-end_gapextend",
            "1",
            "-gapopen",
            "11",
            "-gapextend",
            "1",
            "-blast_evalue",
            "0.003",
            "-norps",
            "T",
            "-treemethod",
            "clust",
        ]

        with aligned_file.open("w") as output_file:
            result = subprocess.run(
                command, stdout=output_file, stderr=subprocess.PIPE, text=True
            )

        if result.returncode == 0:
            print("MSA with COBALT completed successfully.")
            return True
        else:
            print(f"COBALT command failed with error: {result.stderr}")
            return False
